{
    XviD Stats Analyzer
    Copyright (C) 2005 Yuri Khan
    stats-analyzer at gorodok dot net

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
}

unit GraphUnit;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, Contnrs, StdCtrls, Registry, ComCtrls, ColorsUnit;

type
  TFrameInfo = class
  private
    FFrameType: Char;
    FQuantizer: Integer;
    FIntraBlocks: Integer;
    FFrameSize: Integer;
    FInterBlocks: Integer;
    FHeaderSize: Integer;
    FNotCodedBlocks: Integer;
    FFilteredFrameSize: Integer;
  public
    property FrameType: Char read FFrameType;
    property Quantizer: Integer read FQuantizer;
    property IntraBlocks: Integer read FIntraBlocks;
    property InterBlocks: Integer read FInterBlocks;
    property NotCodedBlocks: Integer read FNotCodedBlocks;
    property FrameSize: Integer read FFrameSize;
    property HeaderSize: Integer read FHeaderSize;

    procedure Read(var f: TextFile);
    property FilteredFrameSize: Integer read FFilteredFrameSize
      write FFilteredFrameSize;
  end;

  TProgressEvent = procedure(Sender: TObject; Progress: Integer) of object;

  TGraphForm = class(TForm)
    GraphPanel: TPanel;
    GraphScrollBox: TScrollBox;
    GraphPaintBox: TPaintBox;
    AxisImage: TImage;
    ToolbarPanel: TPanel;
    BitrateThresholdEdit: TLabeledEdit;
    ApplyButton: TButton;
    StatusBar: TStatusBar;
    GotoEdit: TLabeledEdit;
    GoButton: TButton;
    FpsEdit: TLabeledEdit;
    ColorsButton: TButton;
    procedure GraphScrollBoxResize(Sender: TObject);
    procedure GraphPaintBoxPaint(Sender: TObject);
    procedure ApplyButtonClick(Sender: TObject);
    procedure GraphPaintBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure GraphPaintBoxMouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure GraphPaintBoxMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure BitrateThresholdEditEnter(Sender: TObject);
    procedure GotoEditEnter(Sender: TObject);
    procedure GoButtonClick(Sender: TObject);
    procedure ColorsButtonClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    FSelStart: Integer;
    FSelEnd: Integer;
    FOnSelectionChange: TNotifyEvent;
    FPosition: Integer;
    FOnLoadProgress: TProgressEvent;
    FDefaultButton: TButton;
    FFps: Extended;

    function GetFrames(i: Integer): TFrameInfo;
    function GetFrameCount: Integer;
    procedure SetBitrateThreshold(const Value: Integer);
    procedure SetSelEnd(Value: Integer);
    procedure SetSelStart(Value: Integer);
    procedure InvertSelection(DC: HDC);
    procedure SetPosition(Value: Integer);
    procedure InvertPosition(DC: HDC);
    procedure UpdateStatusBar(Value: Integer);
    procedure SetDefaultButton(const Value: TButton);
    procedure UpdateAverageBitrate;
    procedure ScrollView;
  private
    { Private declarations }
    FFrames: TObjectList;
    FGraphWidth: Integer;
    FGraphHeight: Integer;
    FMaxFrameSize: Integer;
    FMaxFrameIndex: Integer;
    FTotalFrameSize: Integer;
    FAverageFrameSize: Integer;
    FBitrateThreshold: Integer;

    FColors: TGraphColors;

    procedure AddFrame(fi: TFrameInfo);
    procedure UpdateGraph;
    function SizeY(s: Integer): Integer;

    property FrameCount: Integer read GetFrameCount;
    property Frames[i: Integer]: TFrameInfo read GetFrames;

    property BitrateThreshold: Integer read FBitrateThreshold write SetBitrateThreshold;

    property DefaultButton: TButton read FDefaultButton write SetDefaultButton;
  public
    { Public declarations }

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    property OnLoadProgress: TProgressEvent read FOnLoadProgress write FOnLoadProgress;
    procedure LoadFile(Filename: string);

    property MaxFrameIndex: Integer read FMaxFrameIndex;

    property SelStart: Integer read FSelStart write SetSelStart;
    property SelEnd: Integer read FSelEnd write SetSelEnd;
    property OnSelectionChange: TNotifyEvent read FOnSelectionChange write FOnSelectionChange;

    property Position: Integer read FPosition write SetPosition;

    procedure SaveSettings(Registry: TRegistry);
    procedure LoadSettings(Registry: TRegistry);
  end;

var
  GraphForm: TGraphForm;

implementation

uses Math;

{$R *.dfm}

function NormDist(x, mu, sigma: Extended): Extended;
begin
  Result := exp(-(sqr(x - mu) / (2 * sqr(sigma)) )) / (sqrt(2 * Pi) * sigma);
end;

{ TGraphForm }

procedure TGraphForm.LoadFile(Filename: string);
const sigma = 30;

var f: TextFile;
  frameInfo: TFrameInfo;
  i, j, pct, oldPct: Integer;
  sum, denom: Extended;
begin
  FileMode := fmOpenRead or fmShareDenyNone;
  AssignFile(f, Filename);
  Reset(f);
  try
    ReadLn(f); // banner
    ReadLn(f); // banner
    ReadLn(f); // empty line

    FMaxFrameSize := 0;
    FFrames.Clear;
    FTotalFrameSize := 0;
    while not EOF(f) do
    begin
      frameInfo := TFrameInfo.Create;
      frameInfo.Read(f);
      AddFrame(frameInfo);
      Inc(FTotalFrameSize, frameInfo.FrameSize);
      if frameInfo.FrameSize > FMaxFrameSize then
      begin
        FMaxFrameSize := frameInfo.FrameSize;
        FMaxFrameIndex := FFrames.Count - 1;
      end;
    end;

    FAverageFrameSize := FTotalFrameSize div FrameCount;
    BitrateThreshold := FTotalFrameSize * 3 div (FrameCount * 2);
    UpdateAverageBitrate;

    oldPct := -1;
    for i := 0 to FrameCount - 1 do
    begin
      if Assigned(FOnLoadProgress) then
      begin
        pct := i * 100 div FrameCount;
        if oldPct <> pct then
        begin
          FOnLoadProgress(Self, pct);
          oldPct := pct;
        end;
      end;

      sum := 0;
      denom := 0;
      for j := Max(-3*sigma, -i) to Min(3*sigma, (FrameCount - 1) - i) do
      begin
        sum := sum + NormDist(j, 0, sigma) * Frames[i + j].FrameSize;
        denom := denom + NormDist(j, 0, sigma);
      end;
      Frames[i].FilteredFrameSize := Round(sum / denom);
    end;

    if Assigned(FOnLoadProgress) then
    begin
      FOnLoadProgress(Self, 100);
    end;

    UpdateGraph;
  finally
    CloseFile(f);
  end;
end;

procedure TGraphForm.AddFrame(fi: TFrameInfo);
begin
  FFrames.Add(fi);
end;

function TGraphForm.SizeY(s: Integer): Integer;
begin
  Result := s * FGraphHeight div FMaxFrameSize;
end;

procedure TGraphForm.UpdateGraph;
var i, right, bottom, top, y: Integer;
begin
  if not Assigned(AxisImage.Picture.Bitmap) then
    AxisImage.Picture.Bitmap := TBitmap.Create;

  AxisImage.Picture.Bitmap.Width := AxisImage.Width;
  AxisImage.Picture.Bitmap.Height := AxisImage.Height;

  GraphPaintBox.Width := FFrames.Count;

  FGraphWidth := FFrames.Count;
  FGraphHeight := AxisImage.Height - 100;

  with AxisImage.Picture.Bitmap, Canvas do
  begin
    right := Width - 1;
    bottom := height - 50;
    top := 50;

    // Background
    Brush.Color := FColors.Background;
    Brush.Style := bsSolid;
    FillRect(ClipRect);

    Pen.Color := FColors.Axes;
    Pen.Style := psSolid;
    Pen.Width := 1;
    MoveTo(right, bottom);
    LineTo(right, top);

    Font.Name := 'MS Shell Dlg';
    Font.Size := 10;
    Font.Color := FColors.Text;
    SetTextAlign(Handle, TA_RIGHT or TA_BOTTOM);
    for i := 0 to (FMaxFrameSize - 1) div 10000 do
    begin
      y := bottom - SizeY(i * 10000);
      MoveTo(right, y);
      LineTo(right - 2, y);
      TextOut(right - 5, y, IntToStr(i * 10000));
    end;
  end;
end;

function TGraphForm.GetFrames(i: Integer): TFrameInfo;
begin
  Result := FFrames[i] as TFrameInfo;
end;

constructor TGraphForm.Create(AOwner: TComponent);
begin
  inherited;
  FFrames := TObjectList.Create;
  FMaxFrameSize := 1;
  FFps := 23.976;
  FColors := TGraphColors.Create;
end;

destructor TGraphForm.Destroy;
begin
  FreeAndNil(FFrames);
  FreeAndNil(FColors);
  inherited;
end;

function TGraphForm.GetFrameCount: Integer;
begin
  Result := FFrames.Count;
end;

procedure TGraphForm.SetBitrateThreshold(const Value: Integer);
begin
  FBitrateThreshold := Value;
  BitrateThresholdEdit.Text := IntToStr(Value);
end;

procedure TGraphForm.GraphScrollBoxResize(Sender: TObject);
begin
  GraphPaintBox.Height := GraphScrollBox.ClientHeight;
  UpdateGraph;
end;

procedure TGraphForm.GraphPaintBoxPaint(Sender: TObject);
var top, bottom, left, right, i, x, y: Integer;
begin
  bottom := AxisImage.Height - 50;
  top := 50;
  left := GraphScrollBox.HorzScrollBar.Position;
  right := GraphScrollBox.HorzScrollBar.Position + GraphScrollBox.Width;

  with GraphPaintBox.Canvas do
  begin
    // Clear
    Brush.Color := FColors.Background;
    Brush.Style := bsSolid;
    FillRect(Rect(left, 0, right, GraphPaintBox.Height));

    // Axis
    Pen.Color := FColors.Axes;
    Pen.Style := psSolid;
    Pen.Width := 1;
    MoveTo(left, bottom);
    LineTo(right, bottom);

    // Size gridlines
    Pen.Color := FColors.Gridlines;
    for i := 0 to (FMaxFrameSize - 1) div 10000 do
    begin
      y := bottom - SizeY(i * 10000);
      MoveTo(left, y);
      LineTo(right, y);
    end;

    // Frame# gridlines and ticks
    Pen.Color := FColors.Gridlines;
    Font.Name := 'MS Shell Dlg';
    Font.Size := 10;
    Font.Color := FColors.Text;
    SetTextAlign(Handle, TA_LEFT or TA_TOP);
    for i := left div 100
    to (right - 1) div 100 do
    begin
      x := 100 * i;
      MoveTo(x, bottom - 1);
      LineTo(x, top);
      TextOut(x, bottom + 5, IntToStr(x));
    end;

    // Data
    for i := left to right - 1 do
    begin
      Pen.Color := FColors.Filtered;
      MoveTo(i, bottom);
      LineTo(i, bottom - SizeY(Frames[i].FilteredFrameSize));

      Pen.Color := FColors.Data;
      MoveTo(i, bottom);
      LineTo(i, bottom - SizeY(Frames[i].FrameSize));
    end;

    // Threshold line
    Pen.Color := FColors.Threshold;
    y := bottom - SizeY(FBitrateThreshold);
    MoveTo(left, y);
    LineTo(right, y);

    // Average line
    Pen.Color := FColors.Average;
    y := bottom - SizeY(FAverageFrameSize);
    MoveTo(left, y);
    LineTo(right, y);

    // Selection
    InvertRect(Handle, Rect(FSelStart, top, FSelEnd, bottom));

    // Position
    InvertRect(Handle, Rect(FPosition, top, FPosition + 1, bottom));
  end;
end;

procedure TGraphForm.ApplyButtonClick(Sender: TObject);
begin
  FBitrateThreshold := StrToInt(BitrateThresholdEdit.Text);
  FFps := StrToFloat(FpsEdit.Text);
  UpdateGraph;
  UpdateAverageBitrate;
  GraphPaintBox.Invalidate;
end;

procedure TGraphForm.SetSelEnd(Value: Integer);
begin
  Value := Max(0, Min(FrameCount, Value));

  FSelEnd := Value;
  GraphPaintBox.Invalidate;
end;

procedure TGraphForm.SetSelStart(Value: Integer);
begin
  Value := Max(0, Min(FrameCount - 1, Value));

  FSelStart := Value;
  GraphPaintBox.Invalidate;
end;

procedure TGraphForm.InvertPosition(DC: HDC);
begin
  InvertRect(DC, Rect(
    FPosition - GraphScrollBox.HorzScrollBar.Position,
    AxisImage.Height - 50 - SizeY(FMaxFrameSize),
    FPosition + 1 - GraphScrollBox.HorzScrollBar.Position,
    AxisImage.Height - 50));
end;

procedure TGraphForm.SetPosition(Value: Integer);
var DC: HDC;
begin
  Value := Max(0, Min(FrameCount - 1, Value));

  DC := GetDC(GraphScrollBox.Handle);
  try
    InvertPosition(DC);
    FPosition := Value;
    InvertPosition(DC);
  finally
    ReleaseDC(GraphScrollBox.Handle, DC);
  end;

  ScrollView;

  UpdateStatusBar(Value);
end;

procedure TGraphForm.ScrollView;
begin
  if (FPosition < GraphScrollBox.HorzScrollBar.Position) or
    (FPosition > GraphScrollBox.HorzScrollBar.Position + GraphScrollBox.Width) then
  begin
    GraphScrollBox.HorzScrollBar.Position :=
      Max(0, Min(FrameCount - GraphScrollBox.Width,
      FPosition - GraphScrollBox.Width div 2));
  end;
end;

procedure TGraphForm.UpdateStatusBar(Value: Integer);
begin
  StatusBar.Panels[0].Text := IntToStr(Value);
  with Frames[Value] do
  begin
    StatusBar.Panels[1].Text := UpperCase(FrameType);
    StatusBar.Panels[2].Text := IntToStr(Quantizer);
    StatusBar.Panels[3].Text := IntToStr(IntraBlocks);
    StatusBar.Panels[4].Text := IntToStr(InterBlocks);
    StatusBar.Panels[5].Text := IntToStr(NotCodedBlocks);
    StatusBar.Panels[6].Text := IntToStr(FrameSize);
    StatusBar.Panels[7].Text := IntToStr(HeaderSize);
  end;
end;

procedure TGraphForm.SetDefaultButton(const Value: TButton);
begin
  if Assigned(FDefaultButton) then
  begin
    FDefaultButton.Default := False;
  end;
  FDefaultButton := Value;
  if Assigned(FDefaultButton) then
  begin
    FDefaultButton.Default := True;
  end;
end;

{ TFrameInfo }

procedure TFrameInfo.Read(var f: TextFile);
begin
  ReadLn(f, FFrameType, FQuantizer,
    FIntraBlocks, FInterBlocks, FNotCodedBlocks,
    FFrameSize, FHeaderSize);
end;

procedure TGraphForm.GraphPaintBoxMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbLeft then
  begin
    FSelStart := X;
    SelEnd := X;
  end;
end;

procedure TGraphForm.InvertSelection(DC: HDC);
begin
  InvertRect(DC, Rect(
    FSelStart - GraphScrollBox.HorzScrollBar.Position,
    AxisImage.Height - 50 - SizeY(FMaxFrameSize),
    FSelEnd - GraphScrollBox.HorzScrollBar.Position,
    AxisImage.Height - 50));
end;

procedure TGraphForm.GraphPaintBoxMouseMove(Sender: TObject;
  Shift: TShiftState; X, Y: Integer);
var DC: HDC;
begin
  if ssLeft in Shift then
  begin
    DC := GetDC(GraphScrollBox.Handle);
    try
      InvertSelection(DC); // Clear selection
      FSelEnd := X;
      UpdateStatusBar(X);
      InvertSelection(DC); // Draw selection
    finally
      ReleaseDC(GraphScrollBox.Handle, DC);
    end;
  end;
end;

procedure TGraphForm.GraphPaintBoxMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var DC: HDC;
begin
  if Button = mbLeft then
  begin
    DC := GetDC(GraphScrollBox.Handle);
    try
      InvertSelection(DC); // Clear selection
    finally
      ReleaseDC(GraphScrollBox.Handle, DC);
    end;
    SelEnd := (X + 65536) mod 65536;
    Position := FSelStart;

    if Assigned(FOnSelectionChange) then
    begin
      FOnSelectionChange(Self);
    end;
  end;
end;

procedure TGraphForm.SaveSettings(Registry: TRegistry);
begin
  try Registry.WriteInteger('Left', Left); except end;
  try Registry.WriteInteger('Top', Top); except end;
  try Registry.WriteInteger('Width', Width); except end;
  try Registry.WriteInteger('Height', Height); except end;
  try Registry.WriteFloat('FPS', FFps); except end;
  FColors.SaveToRegistry(Registry);
end;

procedure TGraphForm.LoadSettings(Registry: TRegistry);
begin
  try Left := Registry.ReadInteger('Left'); except end;
  try Top := Registry.ReadInteger('Top'); except end;
  try Width := Registry.ReadInteger('Width'); except end;
  try Height := Registry.ReadInteger('Height'); except end;
  try FFps := Registry.ReadFloat('FPS'); except end;
  FColors.LoadFromRegistry(Registry);
end;

procedure TGraphForm.BitrateThresholdEditEnter(Sender: TObject);
begin
  DefaultButton := ApplyButton;
end;

procedure TGraphForm.GotoEditEnter(Sender: TObject);
begin
  DefaultButton := GoButton;
end;

procedure TGraphForm.GoButtonClick(Sender: TObject);
var Value: Integer;
begin
  Value := Max(0, Min(FrameCount - 1, StrToInt(GotoEdit.Text)));
  FSelStart := Value;
  SelEnd := Value;
  Position := Value;
  UpdateStatusBar(Value);
  ScrollView;
  if Assigned(FOnSelectionChange) then
  begin
    FOnSelectionChange(Self);
  end;
end;

procedure TGraphForm.UpdateAverageBitrate;
begin
  StatusBar.Panels[8].Text := Format('Average bitrate: %f kbps', [
    FFps{frames/s} * FTotalFrameSize{byte} * 8{bit/byte} /
    (FrameCount{frames} * 1000{bit/kbit})]);
end;

procedure TGraphForm.ColorsButtonClick(Sender: TObject);
begin
  with TColorsForm.Create(Self) do
  try
    Colors.Assign(FColors);
    if Execute then
    begin
      FColors.Assign(Colors);
      UpdateGraph;
      GraphPaintBox.Invalidate;
    end;
  finally
    Free;
  end;
end;

procedure TGraphForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if ActiveControl = GraphScrollBox then
  begin
    case Key of
      VK_LEFT:
        if [ssCtrl] = Shift then
        begin
          Position := Position - 10;
          Key := 0;
        end
        else if [] = Shift then
        begin
          Position := Position - 1;
          Key := 0;
        end;
      VK_RIGHT:
        if [ssCtrl] = Shift then
        begin
          Position := Position + 10;
          Key := 0;
        end
        else if [] = Shift then
        begin
          Position := Position + 1;
          Key := 0;
        end;
    end;
  end;
end;

end.
